from abc import ABCMeta, abstractmethod
from typing import Any
from .meta import MapOf, ArrayOf


class ValidationException(Exception):
    def __init__(self, issues={}):
        """
        Parameters:
        issues: a dictionary of field name: message for validation issues.
        """
        super().__init__()
        self._issues = issues

    @property
    def issues(self):
        """
        Returns a dictionary of field name : message indicating the validation issues.
        """
        return self._issues

    def __str__(self):
        return "ValidationException: " + ", ".join(
            [f"{k}: {v}" for k, v in self._issues.items()]
        )


class Validator(metaclass=ABCMeta):
    """
    Interface for validators
    """

    @abstractmethod
    def validate(self, obj: object, field: str, value: Any) -> str:
        """
        Validates a field value.
        Returns None when validation passes, or a message indicating the issue.
        """
        pass


class TypeValidator(Validator):
    """
    Validates that the value being set is of the type
    declared in the class metadata
    """

    def validate(self, obj, field, value):
        prop_descriptor = getattr(obj.__class__, field)

        if value is None and not prop_descriptor.allow_none:
            return f"{field} must be not null"

        if isinstance(prop_descriptor.type, MapOf):
            if not isinstance(value, dict):
                return f"{field} value '{value}' is not a dictionary-like object"

        if isinstance(prop_descriptor.type, ArrayOf):
            if not isinstance(value, list):
                return f"{field} value '{value}' is not a list-like object"

        if not isinstance(value, prop_descriptor.type):
            return f"{field} value '{value}' is not of type {prop_descriptor.type}"



# vim:ts=4:sw=4:expandtab
import unittest
from podm import JsonObject, Property, Handler, Processor, ArrayOf, MapOf
from collections import OrderedDict
from datetime import datetime
from .common import (
    Entity,
    Company,
    Sector,
    Employee,
    DateTimeHandler,
    TestObject,
    TestObject2,
    Child,
    Parent,
)
from enum import Enum


class Values(Enum):
    VALUE1 = 1
    VALUE2 = 2


class Type2(JsonObject):
    field_1 = Property(type=int, title="Field 1 title")
    field_2 = Property(type=Values, enum_as_str=True)


class Type1(JsonObject):

    field_1 = Property(type=str)
    field_2 = Property(type=Type2)


class TestSchema(unittest.TestCase):
    def test_base_schema(self):
        schema = Company.schema()
        self.assertIn("type", schema)
        self.assertIn("properties", schema)
        self.assertIn("py/object", schema["properties"])
        self.assertIn("const", schema["properties"]["py/object"])
        self.assertEqual(
            "test.common.Company", schema["properties"]["py/object"]["const"]
        )
        self.assertIn("oid", schema["properties"])
        self.assertIn("type", schema["properties"]["oid"])
        self.assertEqual("string", schema["properties"]["oid"]["type"])
        self.assertIn("created", schema["properties"])
        self.assertEqual("object", schema["properties"]["created"]["type"])
        self.assertIn("company-name", schema["properties"])
        self.assertEqual("string", schema["properties"]["company-name"]["type"])
        self.assertIn("description", schema["properties"])
        self.assertEqual("string", schema["properties"]["description"]["type"])

    def test_compound(self):
        import json

        schema = Type1.schema()
        print(json.dumps(schema, indent=4))
        self.assertEqual(schema["type"], "object")
        self.assertEqual(
            schema["properties"]["field_2"]["type"], {"$ref": "#/definitions/Type2"}
        )
        self.assertEqual(1, len(schema["definitions"]))
        self.assertIn("properties", schema["definitions"]["Type2"])
        self.assertIn("field_1", schema["definitions"]["Type2"]["properties"])
        self.assertIn("title", schema["definitions"]["Type2"]["properties"]["field_1"])
        self.assertIn("field_2", schema["definitions"]["Type2"]["properties"])
        self.assertIn("enum", schema["definitions"]["Type2"]["properties"]["field_2"])
        self.assertEqual(
            set(["VALUE1", "VALUE2"]),
            set(schema["definitions"]["Type2"]["properties"]["field_2"]["enum"]),
        )



from unittest import TestCase
from podm import BaseJsonObject, JsonObject, Property, PropertyHandler
import podm.jsonobject
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, Integer, String, ForeignKey
from sqlalchemy.orm import relationship
from sqlalchemy.orm.attributes import InstrumentedAttribute

Base = declarative_base()


class CustomPropertyHandler(PropertyHandler):
    def __init__(self, obj_class, name, definition):
        super().__init__(obj_class, name, definition)

    def set(self, target, value):
        try:
            target.__dict__[self._name] = value
        except KeyError:
            pass

    def get(self, target):
        try:
            return target.__dict__[self._name]
        except KeyError:
            return None

    def json(self):
        return self._name

    def field_name(self):
        self._name


class SomeObject(JsonObject):
    value = Property()


class CustomIntrospector(podm.jsonobject.DefaultIntrospector):
    def get_properties(self, obj_class):

        properties = super().get_properties(obj_class)
        for key, val in obj_class.__dict__.items():
            if isinstance(val, InstrumentedAttribute):
                if key in obj_class.__table__.c:
                    print(key, val, obj_class.__table__.c[key].type)
                    properties[key] = CustomPropertyHandler(obj_class, key, None)

        return properties

    def property_handler_class(self):
        return CustomPropertyHandler


class BaseObject(Base, BaseJsonObject):

    _introspector = CustomIntrospector()

    __abstract__ = True

    id = Column(Integer, primary_key=True)


class House(BaseObject):

    __abstract__ = False
    __tablename__ = "house"

    name = Column(String(50))


class Person(BaseObject):

    """
    Base class for all entities that are stored in SQL Database by using
    SQLAlchemy ORM.

    Attributes:
        id: Entity id, numeric, unique for a given entity type, and autogenerated.
            Note that IDs are populated once session.flush() is performed.
    """

    __abstract__ = False
    __tablename__ = "person"

    name = Column(String(50))
    house_id = Column(Integer, ForeignKey("house.id"))
    house = relationship("House", lazy="joined")

    some_object = Property()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._some_object = None


class TestSQLAlchemy(TestCase):
    def test_serialize(self):

        entity = Person(id=1)
        entity.name = "Carlos"
        entity.some_object = {"1": [SomeObject()]}

        data_dict = entity.to_dict()
        print(data_dict)

        new_instance = Person.from_dict(data_dict)
        print(new_instance.id, new_instance.name)



import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="podm",
    version="0.0.11",
    author="Carlos Descalzi",
    author_email="carlos.descalzi@gmail.com",
    description="A Object-JSON Document mapper",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Carlos-Descalzi/podm",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)
